#include <u.h>
#include <lib9.h>
#include <fmt.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#include "FLAC/stream_decoder.h"

enum {
	Bufsiz = 2048,
};

static int ofd = -1;
static int cfd = -1;
static int vfd = -1;
static int verbose;
static int streaming;

typedef struct Aux Aux;
struct Aux {
	char *file;
	FILE *fp;
	int depth;
	int chans;
	int freq;
};

static int
x_read(FLAC__StreamDecoder *dec, FLAC__byte buf[], unsigned *bytes, void *v)
{
	Aux *a;
	int n, got, want;

	a = v;
	got = 0;
	want = *bytes;

	*bytes = 0;
	while(got < want){
		n = fread(buf +got, 1, want -got, a->fp);
		if(n == 0)
			return FLAC__STREAM_DECODER_READ_STATUS_END_OF_STREAM;
		if(ferror(a->fp))
			return FLAC__STREAM_DECODER_READ_STATUS_ABORT;
		got += n;
	}

	*bytes = got;
	return FLAC__STREAM_DECODER_READ_STATUS_CONTINUE;
}

static int
x_write(FLAC__StreamDecoder *, FLAC__Frame *frame, FLAC__int32 *buf[], void *v)
{
	Aux *a;
	int i, o, s1, s2;
	uchar out[Bufsiz];

	a = v;
	o = 0;
	s1 = (a->depth - 16);
	s2 = (a->depth - 8);

	for(i = 0; i < frame->header.blocksize; i++){
		out[o++] = (unsigned)buf[0][i] >> s1;
		out[o++] = (unsigned)buf[0][i] >> s2;
		if(a->chans == 1){
			out[o++] = (unsigned)buf[0][i] >> s1;
			out[o++] = (unsigned)buf[0][i] >> s2;
		}
		else{
			out[o++] = (unsigned)buf[1][i] >> s1;
			out[o++] = (unsigned)buf[1][i] >> s2;
		}
		if(o >= Bufsiz -4){
			if(write(ofd, out, o) != o)
				sysfatal("write failed");
			o = 0;
		}
	}
	if(write(ofd, out, o) != o)
		sysfatal("write failed");
	return 0;
}

static void
x_error(FLAC__StreamDecoder *, FLAC__StreamDecoderErrorStatus status, void *v)
{
	Aux *a;

	a = v;
	switch(status){
	case FLAC__STREAM_DECODER_ERROR_STATUS_LOST_SYNC:
		print("error 'lost sync'\n");
		break;
	case FLAC__STREAM_DECODER_ERROR_STATUS_BAD_HEADER:
		print("error 'bad frame header'\n");
		break;
	case FLAC__STREAM_DECODER_ERROR_STATUS_FRAME_CRC_MISMATCH:
		print("error 'CRC fail'\n");
		break;
	default:
		print("error '%s'", FLAC__StreamDecoderErrorStatusString[status]);
		break;
	}
	if(! streaming)
		exit(1);

}

static void
x_meta(FLAC__StreamDecoder *, FLAC__StreamMetadata *m, void *v)
{
	Aux *a;
	int n, i;

	a = v;
	switch(m->type){
	case FLAC__METADATA_TYPE_STREAMINFO:
		if(verbose){
			print("enc flac\n");
			print("freq %g\n", (double)m->data.stream_info.sample_rate / 1000.0);
			print("chan %d\n", m->data.stream_info.channels);
		}
	
		n = m->data.stream_info.sample_rate;
		if(vfd < 0)
			fprint(2, "warning: audio sample rate is %d Hz\n", n);
		else{
			fprint(vfd, "speed %d\n", n);
			fprint(cfd, "rate %d\n", n);
		}
	
		a->chans = m->data.stream_info.channels;
		a->depth = m->data.stream_info.bits_per_sample;
		a->freq = m->data.stream_info.sample_rate;
		break;

	case FLAC__METADATA_TYPE_VORBIS_COMMENT:
		if(verbose){
			print("comment VENDOR=%s\n", m->data.vorbis_comment.vendor_string);
			for(i = 0; i < m->data.vorbis_comment.num_comments; i++)
				print("comment %.*s\n", m->data.vorbis_comment.comments[i].length,
					m->data.vorbis_comment.comments[i].entry);
		}
		break;
	default:
		break;
	}
}


int
play(FILE *fp, char *file)
{
	Aux aux;
	FLAC__StreamDecoder *dec;

	aux.fp = fp;
	aux.file = file;
	dec = FLAC__stream_decoder_new();

	FLAC__stream_decoder_set_read_callback(dec, x_read);
	FLAC__stream_decoder_set_write_callback(dec, x_write);
	FLAC__stream_decoder_set_error_callback(dec, x_error);
	FLAC__stream_decoder_set_metadata_callback(dec, x_meta);

	FLAC__stream_decoder_set_client_data(dec, &aux);

	FLAC__stream_decoder_init(dec);
	FLAC__stream_decoder_process_until_end_of_stream(dec);
	FLAC__stream_decoder_finish(dec);

	return 0;
}

static void
usage(void)
{
	fprint(2, "%s [-s] [-v] [-o outfile] [file...]\n", argv0);
	exit(1);
}

void
main(int argc, char **argv)
{
	int i, fd;
	FILE *fp;
	char *p, *outfile;

	outfile = nil;
	ARGBEGIN{
	case 'v':
		verbose++;
		break;
	case 's':
		streaming++;
		break;
	case 'o':
		outfile = EARGF(usage());
		break;
	default:
		usage();
	}ARGEND
	
	if(outfile){
		if((ofd = creat(outfile, 0666)) < 0)
			sysfatal("creat %s: %r", outfile);
	}else{
		if((ofd = open("/dev/audio", O_WRONLY)) < 0)
			sysfatal("open /dev/audio: %r");
		vfd = open("/dev/volume", O_WRONLY);
		cfd = open("/dev/audioctl", O_WRONLY);
	}

	if(argc == 0){
		play(stdin, "stdin");
	}else{
		for(i=0; i<argc; i++){
			if((fp = fopen(argv[i], "r")) == nil)
				sysfatal("fopen %s: %r", argv[i]);
			play(fp, argv[i]);
			fclose(fp);
		}
	}
	exit(0);
}
