/*
 * Decode ADTS (audio data transport stream)
 * We try to 
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <lib9.h>
#include <fmt.h>

#include "neaacdec.h"

int verbose;
int tollerant;

int ofd = -1;
int vfd = -1;
int cfd = -1;

enum {
	Bufsiz = 8192,
	AtdsHdrLen = 7,		/* adts hearder is 56 bits */
};

static char *Wrap[] = {
	[RAW] 		"raw",
	[ADTS]		"adts",
	[ADIF]		"adif"
};

char *Prof[] = {
	[MAIN]		"main",
	[LC]		"lc",
	[SSR]		"ssr",
	[LTP]		"ltp",
	[HE_AAC]	"he",
	[ER_LC]		"er-lc",
	[ER_LTP]	"er-ltp",
	[LD]		"ld",
	[DRM_ER_LC]	"drm-er-lc"
};

/*
 * The faad2 decoder is not robust to malformed data as may be seen if you
 * jump into the middle of an audio stream, so we explciitly lock onto the
 * frame structure before offering the stream to faad2.
 */
typedef struct Hdr Hdr;
struct Hdr {
	uvlong bytes;		/* byte offset in stream */
	uvlong buf;			/* current input buffer */
	uint prev;			/* fixed portion of previous header found */
	vlong next;			/* expected offset of next header */
	vlong last;			/* offset of last plausable header */
};

static uint
bits(Hdr *h, int off, int len)
{
	return (h->buf >> off) & ((1 << len) -1);
}

static int
validhdr(Hdr *h)
{
	uint fixed;
	int frmlen;

	if(h->next != 0 && h->bytes != h->next)	/* locked, searching? */
		return h->next - h->bytes;

	fixed = bits(h, 28, 28);
	if(fixed != h->prev && h->bytes == h->next){
		h->prev = fixed;
		goto fail;
	}

	if(bits(h, 44, 12) != 0xfff)	/* bad preamble */
		goto fail;
	if(bits(h, 41, 2) != 0)			/* bad layer */
		goto fail;
	if(bits(h, 38, 2) == 0)			/* bad profile */
		goto fail;
	if(bits(h, 34, 4) == 15) 		/* bad sample rate */
		goto fail;
	if(bits(h, 33, 1) != 0)			/* bad private flag */
		goto fail;
	if(bits(h, 29, 1) != 0)			/* bad origin */
		goto fail;
	if(bits(h, 28, 1) != 0)			/* bad home */
		goto fail;
	if(bits(h, 23, 3) >= 8)			/* bad channel spec */
		goto fail;

	frmlen = bits(h, 13, 13);
	if(h->next == 0){				/* was unlocked? */
//		fprint(2, "framelock: skipped %lld\n", h->bytes - h->last);
		h->next = h->bytes + frmlen;
		return -1;					/* valid header, possible lock */
	}

	h->last = h->bytes;
	h->next = h->bytes + frmlen;
	return 0;						/* am locked, got valid header */

fail:
	if(h->bytes == h->next)			/* where we locked? */
		h->next = 0;
	return -1;						/* lost lock */
}


int
framelock(int fd)
{
	uchar c;
	int nxthdr;
	Hdr fh;

	memset(&fh, 0, sizeof(fh));
	do{
		if(read(fd, &c, 1) < 0)
			return -1;

		fh.bytes++;
		fh.buf = (fh.buf << 8) | (c & 0xff);

		nxthdr = validhdr(&fh);

		/*
		 * if we are locked and we are now 
		 * AtdsHdrLen bytes before the end of the header
		 * (i.e. at the start of the next header),
		 * we can switch to passing data to
		 * audio decoder.
		 */
	}while(nxthdr != AtdsHdrLen);
	return 0;
}


int
fillbuf(int fd, uchar *buf, int off, int want)
{
	int n;

	while(want - off > (want / 16)){
		if((n = read(fd, buf+off, want-off)) <= 0)
			break;
		off += n;
	}
	return off;
}

void
prinfo(NeAACDecFrameInfo *i)
{
	double rate, freq;

	static ulong num = 0;
	static ulong sum = 0;
	static int header_type = -1;
	static int channels = -1;
	static double orate = -1;
	static int samplerate = -1;
	static int object_type = -1;

	if(header_type != i->header_type)
		print("wrap adts\n", Wrap[i->header_type]);

	freq = (double)i->samplerate / 1000.0;
	if(i->samplerate != samplerate)
		print("freq %g\n", freq);

	sum += i->bytesconsumed - AtdsHdrLen;
	num++;
	if(num > 0x7fffffff){				/* prevent overflow */
		sum /= 8;
		num /= 8;
	}
	if(num > 100){					/* allow the average bitrate to stabalise */
		rate = 8 * (sum / (num * 1000.0)) * (i->samplerate / 1024.0);
		if(fabs(rate - orate) > 0.9)
			print("rate %d\n", (int)(rate + 0.5));
	}

	if(i->channels != channels)
		switch(i->channels){
		case 1:
			print("chan Mono\n");
			break;
		case 2:
			print("chan Stereo\n");
			break;
		default:
			print("chan %d\n", i->channels);
			break;
		}

	if(i->object_type != object_type)
		print("enc aac-%s\n", Prof[i->object_type]);

	orate = rate;
	channels = i->channels;
	samplerate = i->samplerate;
	header_type = i->header_type;
	object_type = i->object_type;
}

int
play(int fd, char *file)
{
 	ulong rate;
	ushort *dec;
	NeAACDecHandle aac;
 	int i, n, o, len, sz;
	NeAACDecFrameInfo inf;
	NeAACDecConfigurationPtr cfg;
	uchar chans, inp[Bufsiz], out[Bufsiz];

	if((len = fillbuf(fd, inp, 0, Bufsiz)) <= 0)
		return 0;

	aac = NeAACDecOpen();
	cfg = NeAACDecGetCurrentConfiguration(aac);
	cfg->defSampleRate = 44100;
	cfg->defObjectType = LC;
	cfg->outputFormat = FAAD_FMT_16BIT;
	cfg->downMatrix = 0;
	cfg->useOldADTSFormat = 0;
	cfg->dontUpSampleImplicitSBR = 0;
	NeAACDecSetConfiguration(aac, cfg);

	if((n = NeAACDecInit(aac, inp, len, &rate, &chans)) < 0)
		sysfatal("NAACDecInit() failed");
	memmove(inp, inp+n, len-n);
	len -= n;

	if(vfd < 0)
		fprint(2, "warning: audio sample rate is %d Hz\n", rate);
	else{
		fprint(vfd, "speed %d\n", rate);
		fprint(cfd, "rate %d\n", rate);
	}

	o = 0;
	while((len = fillbuf(fd, inp, len, Bufsiz)) > 0){

		if((dec = NeAACDecDecode(aac, &inf, inp, len)) == nil || inf.error != 0)
				sysfatal("%s - cannot decode AAC audio, %s", file,
					NeAACDecGetErrorMessage(inf.error));

		if(verbose)
			prinfo(&inf);

		for(i = 0; i < inf.samples; i += inf.channels){
			out[o++] = dec[i];
			out[o++] = dec[i] >> 8;

			if(inf.channels == 1){
				out[o++] = dec[i];
				out[o++] = dec[i] >> 8;
			}
			else{
				out[o++] = dec[i+1];
				out[o++] = dec[i+1] >> 8;
			}

			if(o >= Bufsiz -4){
				if(write(ofd, out, o) != o)
					sysfatal("write failed");
				o = 0;
			}
		}

		memmove(inp, inp+inf.bytesconsumed, len-inf.bytesconsumed);
		len -= inf.bytesconsumed;
	}

	if(write(ofd, out, o) != o)
		sysfatal("write failed");

	NeAACDecClose(aac);
	free(out);
	return 0;
}


void
usage(void)
{
	fprint(2, "%s [-o outfile] [file...]\n", argv0);
	exit(1);
}

int
main(int argc, char *argv[])
{
	int i, fd;
	char *outfile;

	outfile = NULL;
	ARGBEGIN{
	case 's':			/* input is a stream, tollerate errors until we are locked */
		tollerant++;
		break;
	case 'v':
		verbose++;
		break;
	case 'o':
		outfile = EARGF(usage());
		break;
	default:
		usage();
	}ARGEND;

	if(outfile){
		if((ofd = creat(outfile, 0644)) < 0)
			sysfatal("%s cannot create, %r", outfile);
	}else{
		if((ofd = open("/dev/audio", O_WRONLY)) < 0)
			sysfatal("/dev/audio cannot open, %r");
		vfd = open("/dev/volume", O_WRONLY);
		cfd = open("/dev/audioctl", O_WRONLY);
	}

	if(argc == 0){
		if(tollerant == 0 || framelock(0) == 0)
			play(0, "standard input");
	}
	else{
		for(i=0; i<argc; i++){
			if((fd = open(argv[i], O_RDONLY)) < 0)
				sysfatal("%s cannot open, %r", argv[i]);
			if(tollerant == 0 || framelock(fd) == 0)
				play(fd, argv[i]);
			close(fd);
		}
	}
	return 0;
}
