/*
 * Copyright (c) 2011 Apple Inc. All rights reserved.
 *
 * @APPLE_APACHE_LICENSE_HEADER_START@
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * @APPLE_APACHE_LICENSE_HEADER_END@
 */

/*
	File:		ALACDecoder.h
*/

#ifndef _ALACDECODER_H
#define _ALACDECODER_H


#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


/*! 
 * The values of these enumerations must match the values used internally by the library.
 */
enum
{
	alac_ok = 0, /* All is good. No error. */
	alac_param_error = -50, /* A parameter passed to the function was invalid in some way. */
	alac_alloc_error = -108 /* Memory allocation failed. */
};


/*!
 * Opaque type for the ALAC decoder object.
 */
typedef void* ALACDecoder;


/*! Create a new ALAC decoder object.
 * @return If memory allocation fails during object creation, the returned value will be @c NULL, otherwise the returned object will be a valid ALAC decoder.
 */
ALACDecoder alac_create_decoder(void);


/*! Initialize the ALAC decoder using the magic cookie found in an .m4a file.
 * @param [in] decoder An ALAC decoder object created with @ref alac_create_decoder.
 * @param [in] magic_cookie Pointer to the buffer returned by @ref mp4ff_get_decoder_config. @b Must not be @c NULL.
 * @param [in] magic_cookie_size The value returned from @ref mp4ff_get_decoder_config. @b Must be 24 or 48.
 * @retval alac_ok The magic cookie was valid and the decoder is ready to decode audio samples.
 * @retval alac_param_error The magic cookie was invalid and the decoder is in an undefined state.
 * @retval alac_alloc_error Allocating memory failed and the decoder is in an undefined state.
 */
int32_t alac_decoder_init( ALACDecoder dec, const uint8_t* magic_cookie, uint32_t magic_cookie_size );


/*! Get the number of sample bits used.
 * @param [in] decoder An ALAC decoder object created with @ref alac_create_decoder and initialized with @ref alac_decoder_init.
 * @return 16, 24 or 32.
 */
int32_t alac_get_sample_bits( const ALACDecoder dec );


/*! Get the number of bytes used by each decoded sample in the ALAC file.
 * @param [in] decoder An ALAC decoder object created with @ref alac_create_decoder and initialized with @ref alac_decoder_init.
 * @return 2 or 4.
 */
int32_t alac_get_bytes_per_sample( const ALACDecoder dec );


/*! Get the frame size used in the ALAC stream.
 * @param [in] decoder An ALAC decoder object created with @ref alac_create_decoder and initialized with @ref alac_decoder_init.
 * @return The number of samples in each compressed frame.
 */
int32_t alac_get_frame_size( const ALACDecoder dec );


/*! Get the number of audio channels in the ALAC stream.
 * @param [in] decoder An ALAC decoder object created with @ref alac_create_decoder and initialized with @ref alac_decoder_init.
 * @return The number of channels of audio in the ALAC file.
 */
int32_t alac_get_channel_count( const ALACDecoder dec );


/*!
 * @param [in] decoder An ALAC decoder object created with @ref alac_create_decoder and initialized with @ref alac_decoder_init.
 * @param [in] in_buffer Pointer to the buffer retrieved by calling @ref mp4ff_read_sample
 * @param [in] in_buffer_size Count of bytes in @c in_buffer.
 * @param [out] sampleBuffer Buffer to write decoded samples. @b Must be at least
 * @param [in] numSamples Number of samples the sampleBuffer can hold.
 * @param [in] numChannels How many channels to decode.
 * @param [out] outNumSamples How many frames of audio decoded to @c sampleBuffer.
 * @retval alac_ok
 * @retval alac_param_error
 * @retval alac_alloc_error
 */
int32_t alac_decode_frame( ALACDecoder dec, uint8_t* in_buffer, uint32_t in_buffer_size, uint8_t* sampleBuffer, uint32_t numSamples, uint32_t numChannels, uint32_t* outNumSamples );


/*! Release all resources used by the ALAC decoder.
 * @param [in] decoder An ALAC decoder object created with @ref alac_create_decoder.
 */
void alac_destroy_decoder( ALACDecoder dec );

#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* _ALACDECODER_H */
