/*
 * Simple mp3 player.  Derived from libmad's example minimad.c.
 */
#include <u.h>
#include <libc.h>
#include "libmad/mad.h"

static int rate = -1;
static int tollerant;
static int verbose;
static char *outfile;
static int ofd = -1;
static int vfd = -1; 		/* /dev/volume */
static int cfd = -1; 		/* /dev/audioctl */

typedef struct Dither Dither;
struct Dither
{
	mad_fixed_t error[3];
	mad_fixed_t random;
};

#define PRNG(x) (((x)*0x19660dL + 0x3c6ef35fL) & 0xffffffffL)

enum
{
	FracBits = MAD_F_FRACBITS,
	OutBits = 16,
	Round = 1 << (FracBits+1-OutBits-1),	// sic
	ScaleBits = FracBits + 1 - OutBits,
	LowMask  = (1<<ScaleBits) - 1,
	Min = -MAD_F_ONE,
	Max = MAD_F_ONE - 1,
};

typedef struct State State;
struct State {
	int fd;
	uchar inpbuf[8192];		/* seems that it needs to be at least 2048 */
	uchar outbuf[8192];		/* > 2881 + MAD_BUFFER_GUARD */
	uchar *p;
	Dither dither;
	char *file;
	int offset;
};

int
audiodither(mad_fixed_t v, Dither *d)
{
	int out;
	mad_fixed_t random;

	/* noise shape */
	v += d->error[0] - d->error[1] + d->error[2];
	d->error[2] = d->error[1];
	d->error[1] = d->error[0] / 2;
	
	/* bias */
	out = v + Round;
	
	/* dither */
	random = PRNG(d->random);
	out += (random & LowMask) - (d->random & LowMask);
	d->random = random;
	
	/* clip */
	if(out > Max){
		out = Max;
		if(v > Max)
			v = Max;
	}else if(out < Min){
		out = Min;
		if(v < Min)
			v = Min;
	}
	
	/* quantize */
	out &= ~LowMask;
	
	/* error feedback */
	d->error[0] = v - out;
	
	/* scale */
	return out >> ScaleBits;
}

static enum mad_flow
input(void *data, struct mad_stream *stream)
{
	int fd, rc, got, n;
	State *s = data;

	got = stream->bufend - stream->next_frame;
	memmove(s->inpbuf, stream->next_frame, got);

	rc = MAD_FLOW_CONTINUE;
	while(sizeof(s->inpbuf)-got > sizeof(s->inpbuf)/10){
		n = read(s->fd, s->inpbuf+got, sizeof(s->inpbuf)-got);
		if(n < 0)
			sysfatal("read failed, %r");
		if(n == 0){
			rc = MAD_FLOW_STOP;
			break;
		}
		got += n;
	}

	mad_stream_buffer(stream, s->inpbuf, got);
	return rc;
}

static enum mad_flow
header(void *data, struct mad_header *hdr)
{
	double enc;
	static int orate = -1;
	static int ofreq = -1;
	static int omode = -1;
	static int oemph = -1;
	static double oenc = -1;
	char *mode[] = { "Mono", "Dual", "Joint", "Stereo" };
	char *emph[] = { "none", "50/15μs", "unknown", "itu-j17" };

 	if(hdr->samplerate != rate){
		rate = hdr->samplerate;
		if(vfd < 0)
			print("warning: audio sample rate is %d Hz\n", rate);
		else{
			fprint(vfd, "speed %d\n", rate);
			fprint(cfd, "rate %d\n", rate);		/* for annoying ac97 driver, till its fixed */
		}
	}

	if(verbose){
		enc = hdr->layer;
		if(hdr->flags & MAD_FLAG_MPEG_2_5_EXT)
			enc = 2.5;

 		if(oenc != enc)
			print("enc mp%g\n",enc);
		if(orate != hdr->bitrate)
			print("rate %g\n", hdr->bitrate / 1024.0);
		if(ofreq != hdr->samplerate)
			print("freq %g\n", hdr->samplerate / 1000.0);
		if(omode != hdr->mode)
			print("chan %s\n", mode[hdr->mode]);
		if(oemph != hdr->emphasis)
			print("emph %s\n", emph[hdr->emphasis]);
	
		oenc = enc;
		orate = hdr->bitrate;
		ofreq = hdr->samplerate;
		omode = hdr->mode;
		oemph = hdr->emphasis;
	}

	return MAD_FLOW_CONTINUE;
}

static enum mad_flow
output(void *data, struct mad_header const* header, struct mad_pcm *pcm)
{
	int i, v;
	State *s = data;
	mad_fixed_t const *left, *right;

	left = pcm->samples[0];
	right = pcm->samples[1];
	for(i=0; i<pcm->length; i++){
		v = audiodither(*left++, &s->dither);
		*s->p++ = v;
		*s->p++ = v>>8;
		if(pcm->channels == 1){
			*s->p++ = v;
			*s->p++ = v>>8;
		}
		else{
			v = audiodither(*right++, &s->dither);
			*s->p++ = v;
			*s->p++ = v>>8;
		}
		if(s->p - s->outbuf >= sizeof(s->outbuf)-4){
			if(write(ofd, s->outbuf, s->p - s->outbuf) != s->p - s->outbuf)
				sysfatal("write error, %r");
			s->p = s->outbuf;
		}
	}
	return MAD_FLOW_CONTINUE;
}

static enum mad_flow
error(void *data, struct mad_stream *stream, struct mad_frame *frame)
{
	State *s = data;

	if(tollerant && MAD_RECOVERABLE(stream->error))
		return MAD_FLOW_CONTINUE;

	sysfatal("%s:#%lld: %s",
		s->file, s->offset - (stream->bufend - stream->next_frame),
		mad_stream_errorstr(stream));
	return 0;
}

void
play(int fd, char *file)
{
	State state;
	struct mad_decoder decoder;
	
	memset(&state, 0, sizeof(state));
	state.file = file;
	state.offset = 0;
	state.p = state.outbuf;

	mad_decoder_init(&decoder, &state, input, header, nil, output, error, nil);
	mad_decoder_run(&decoder, MAD_DECODER_MODE_SYNC);

	if(write(ofd, state.outbuf, state.p - state.outbuf) != state.p - state.outbuf)
		sysfatal("write error, %r");
	state.p = state.outbuf;

	mad_decoder_finish(&decoder);
}

void
usage(void)
{
	fprint(2, "usage: %s [-s] [-v] [-o outfile] [file...]\n", argv0);
	exits("usage");
}

void
main(int argc, char **argv)
{
	int i, fd;
	char *p;

	ARGBEGIN{
	case 's':			/* input is a stream, tollerate recoverable errors */
		tollerant++;
		break;
	case 'v':
		verbose++;
		break;
	case 'o':
		outfile = EARGF(usage());
		break;
	default:
		usage();
	}ARGEND
	
	if(outfile){
		if((ofd = create(outfile, OWRITE, 0666)) < 0)
			sysfatal("create %s: %r", outfile);
	}else{
		if((ofd = open("/dev/audio", OWRITE)) < 0)
			sysfatal("open /dev/audio: %r");
		vfd = open("/dev/volume", OWRITE);
		cfd = open("/dev/audioctl", OWRITE);
	}

	if(argc == 0){
		play(0, "stdin");
	}else{
		for(i=0; i<argc; i++){
			if((fd = open(argv[i], OREAD)) < 0)
				sysfatal("open %s: %r", argv[i]);
			play(fd, argv[i]);
			close(fd);
		}
	}
	exits(0);
}
