#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>
#include "dat.h"
#include "fns.h"

/*
 * Note: we ignore any information we can work out as it may
 * be (often is) wrong:
 *
 * - disconunity - we always check the ContentType HTTP header or, 
 *	 filename extension if none, to chose a codec. If it changes we
 *	 open a new codec. All codecs reset themselves if the input
 *	 changes significantly (bitrate, numchans etc).
 *
 * also beware spurions carriage returns and linefeeds, and pssible
 * missing carriage return or linefeed on last line of the file.
 */

typedef struct Aux Aux;
typedef struct Record Record;

struct Aux {
	Chunk *chunk;		/* chunk currently being built */
	Chunk **tail;		/* tail of chunk list */
};

struct Record {
	char *name;
	int (*func)(Playlist *, char *);
};

static int
sequence(Playlist *pl, char *args)
{
	pl->seqbase = strtoull(args, nil, 10);
	return 0;
}

static int
byterange(Playlist *pl, char *args)
{
	Aux *a;
	vlong len;
	Chunk *ch;

	a = pl->aux;
	if(a->chunk == nil)
		a->chunk = emalloc(sizeof(Chunk));
	ch = a->chunk;

	len = strtoll(args, &args, 10);
	if(*args++ != '@')
		sysfatal("%q badly formatted byterange", args);
	ch->start = strtoll(args, nil, 10);
	ch->end = ch->start+len -1;				/* -1 due to gaps and telegpah poles */
	return 0;
}

static int
stream(Playlist *pl, char *args)
{
	Aux *a;
	int i, n;
	Chunk *ch;
	char *fld[16];

	a = pl->aux;
	if(a->chunk == nil)
		a->chunk = emalloc(sizeof(Chunk));
	ch = a->chunk;

	n = gettokens(args, fld, nelem(fld), ",");
	for(i = 0; i < n; i++)
		if(cistrncmp(fld[i], "BANDWIDTH=", 10) == 0)
			ch->rate = strtol(fld[i]+10, nil, 10);
	return 0;
}

static int
extinf(Playlist *pl, char *args)
{
	Aux *a;
	Chunk *ch;

	pl->isaudio = 1;

	a = pl->aux;
	if(a->chunk == nil)
		a->chunk = emalloc(sizeof(Chunk));
	ch = a->chunk;

	ch->secs = strtod(args, &args);
	if(*args == ',')
		args++;
	ch->title = estrdup(skipws(args));
	return 0;
}

static int
listtype(Playlist *pl, char *args)
{
	if(cistrcmp(args, "vod") == 0)
		pl->immutable = 1;
	return 0;
}

static int
endlist(Playlist *pl, char *)
{
	pl->immutable = 1;
	return 0;
}

Record Records[] = {
	{ "#EXT-X-STREAM-INF:",			stream },
	{ "#EXT-X-MEDIA-SEQUENCE:",		sequence },
	{ "#EXTINF:",					extinf },
	{ "#EXT-X-PLAYLIST_TYPE:",		listtype },
	{ "#EXT-X-ENDLIST:",			endlist },
	{ "#EXT-X-BYTERANGE:",			byterange },
	{ nil,							nil },
};


Playlist *
m3u8(char *url, int fd)
{
	int n;
	Aux *a;
	Chunk *ch;
	Record *r;
	Biobuf bin;
	Playlist *pl;
 	char buf[8192], *base, *rec;

	base = baseurl(url);
	Binit(&bin, fd, OREAD);

	pl = emalloc(sizeof(Playlist));

	a = emalloc(sizeof(Aux));
	pl->aux = a;
	a->tail = &pl->chunks;

	while(Bgetline(&bin, buf, sizeof(buf)) != -1){
		rec = skipws(buf);
		if(*rec == 0)							/* skip blank lines */
			continue;

		if(cistrncmp(rec, "#EXT", 4) == 0){		/* structured comment */
			for(r = Records; r->name; r++){
				n = strlen(r->name);
				if(strncmp(rec, r->name, n) == 0){
					r->func(pl, rec + n);
					break;
				}
			}
			continue;
		}

		if(*rec != '#'){						/* relative filename or a URL */
			if(a->chunk == nil)
				a->chunk = emalloc(sizeof(Chunk));
			ch = a->chunk;
			a->chunk = nil;

			ch->url = absurl(base, rec);

			*(a->tail) = ch;
			a->tail = &ch->next;
			continue;
		}

		if(dbg("m3u8"))
			fprint(2, "m3u8() skip %q\n", rec);
	}
	Bterm(&bin);
	free(base);

	return pl;
}

