#include <u.h>
#include <libc.h>
#include <bio.h>
#include "dat.h"
#include "fns.h"

char *Debug;

typedef struct Mime Mime;
static struct Mime {
	char *type;
	char *ext;
	Playlist *(*func)(char *url, int fd);
	char *codec;
} Mimetypes[] = {
	{ "audio/x-scpls", 					"pls",		pls },
	{ "application/pls+xml", 			"pls",		pls },				/* vpr.net weirdness */
	{ "audio/x-mpegurl",				"m3u8",		m3u8 },				/* ascii */
	{ "audio/mpegurl",					"m3u8",		m3u8 },				/* http://www.rozhlas.cz/audio/download/jazz_high_ogg.m3u */
	{ "application/x-mpegurl",			"m3u8",		m3u8 },				/* utf8 */
	{ "application/vnd.apple.mpegurl",	"m3u8",		m3u8 },				/* HLS stream */

	{ "audio/aacp", 					"aac",		nil,		"/bin/audio/adtsdec" },	/* AAC in MP2 TS */
	{ "audio/aac", 						"aac",		nil, 		"/bin/audio/adtsdec" },
	{ "audio/3gpp",						"3gpp",		nil, 		"/bin/audio/adtsdec" },
	{ "audio/3gpp2",					"3gpp2",	nil, 		"/bin/audio/adtsdec" },

	{ "audio/mpeg", 					"mpeg",		nil,		"/bin/audio/mad" },		/* MP1, MP2, or MP3 */
	{ "audio/x-mpeg", 					"mpeg",		nil,		"/bin/audio/mad" },
	{ "audio/mp2",						"mp2",		nil,		"/bin/audio/mad" },
	{ "audio/mp3",						"mp3",		nil,		"/bin/audio/mad" },

	{ "audio/x-aiff",					"aiff",		nil,		"/bin/audio/aiffdec" },

	{ "audio/mp4a-latm",				"m4a",		nil,		"/bin/audio/m4adec" },	/* AAC or ALAC in an MP4 wrapper */

	{ "audio/x-wav",					"wav",		nil,		"/bin/audio/wavdec" },

	{ "audio/x-flac",					"flac",		nil,		"/bin/audio/flacdec" },
	{ "application/ogg",				"ogg",		nil,		"/bin/audio/flacdec" },

	{ "video/MP2T", 					"ts",		nil,		"/bin/audio/mpgtsdec" },	/* HLS */

	{ nil, nil, nil }
};

static int
codec(char *path)
{
	char *arg[10];
	int i, narg, fd[2];

	narg = 0;
	arg[narg++] = path;
	arg[narg++] = "-s";
	arg[narg++] = "-v";
	arg[narg] = nil;

	if(pipe(fd) < 0){
		fprint(2, "pipe failed %r");
		return -1;
	}

	switch(fork()){
	case -1:
		fprint(2, "fork failed %r");
		close(fd[0]);
		close(fd[1]);
		return -1;
	case 0:
		dup(fd[0], 0);
		close(fd[0]);
		close(fd[1]);
		for(i = 3; i < 63; i++)
			close(i);
		setprio(getpid(), 16);
		exec(path, arg);
		sysfatal("exec %s failed %r", path);
	default:
		close(fd[0]);
		break;
	}
	return fd[1];
}

static int
stream(Sess *s, int conn, int net, char *path)
{
	char buf[8192];
	int got, n, len, metaint;

	metaint = icyhdr(s, conn);

	/* start a codec only on changes */
	if(strcmp(s->decpath, path) != 0){
		strcpy(s->decpath, path);
		if(s->dec != -1)
			close(s->dec);
		if((s->dec = codec(path)) == -1)
			return -1;
	}

	if(metaint == -1){
		while(1){
			if((n = read(net, buf, sizeof(buf))) < 1)
				break;
			if(write(s->dec, buf, n) != n)
				break;
		}
	}
	else{
		while(1){
			for(got = 0; got < metaint; got += n){
	
				len = sizeof(buf);
				if(metaint - got < len)
					len = metaint - got;
	
				if((n = read(net, buf, len)) < 1)
					goto done;
				if(write(s->dec, buf, n) != n)
					goto done;
			}
			if(icyinband(net) < 0)
				break;
		}
	}

done:
	return 0;
}

static Mime *
mimetype(Sess *s, int conn, char *url)
{
	Mime *m;
	char *p, *ext, buf[1024];

	*buf = 0;
	if(httphdr(s, conn, "contenttype", buf, sizeof(buf)) == 0){
		/* some sites add other semicolon delimited fields */
		if((p = strchr(buf, ';')) != nil)
			*p = 0;

		for(m = Mimetypes; m->type; m++)
			if(cistrcmp(m->type, buf) == 0)
				return m;

		werrstr("'%s' ContentType unknown", buf);
	}
	else{	/* No contentype, try and match by filename.ext (ugh!) */
		if((ext = strrchr(url, '.')) != nil)
			for(m = Mimetypes; m->type; m++)
				if(cistrcmp(m->ext, ext +1) == 0)		/* +1 to step over '.' */
					return m;

		werrstr("'%s' extension unknown, no ContentType", ext);
	}

	return nil;
}

int
play(Sess *s, int conn, char *url, vlong start, vlong end)
{
	Mime *m;
	Chunk *ch;
	vlong seq;
	int fd, rc;
	Playlist *pl;

	if((fd = httpdial(s, conn, url, start, end)) < 0){
		werrstr("%s, %r", url);
		return -1;
	}

	if((m = mimetype(s, conn, url)) == nil){
		close(fd);
		return -1;
	}

	if(dbg("mime"))
		fprint(2, "%s audio stream\n", m->type);

	if(m->codec && *m->codec){
		rc = stream(s, conn, fd, m->codec);
		close(fd);
		return rc;
	}

	pl = (*m->func)(url, fd);

	close(fd);

	if(pl == nil){
		werrstr("playlist empty");
		return -1;
	}

	if(s->seqno == 0)	/* first we have seen, also, maybe always zero */
		s->seqno = pl->seqbase;

	if(pl->immutable)
		rc = 1;			/* static playlist, exit when complete */
	else
		rc = 0;			/* dynamic playlist, keep re-reading */

	if(dbg("seq"))
		fprint(2, "seq newlist want=%lld base=%lld\n", s->seqno, pl->seqbase);

	if(pl->isaudio){
		seq = pl->seqbase;
		ch = pl->chunks;

		for(; ch && seq < s->seqno; ch = ch->next){
			if(dbg("dupe"))
				fprint(2, "seq dupe want=%lld got=%lld\n", s->seqno, seq);
			seq++;
		}
		if(seq > s->seqno)
			if(dbg("loss"))
				fprint(2, "seq loss want=%lld got=%lld\n", s->seqno, seq);
		s->seqno = seq;

		for(; ch; ch = ch->next){
			if(dbg("seq"))
				fprint(2, "seq play got=%lld\n", s->seqno);
			if(play(s, conn+1, ch->url, ch->start, ch->end) == -1)
				rc = -1;
			s->seqno++;
		}
	}
	else{
		for(ch = pl->chunks; ch; ch = ch->next){
			while((rc = play(s, conn+1, ch->url, ch->start, ch->end)) == 0){
				if(pl->immutable)
					break;
				if(dbg("seq"))
					fprint(2, "seq list %s\n", ch->url);
			}
				
		}
	}

	freeplaylist(pl);
	return rc;
}

static void
usage(void)
{
	fprint(2, "usage: %s url\n", argv0);
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int rc, n, i;
	Sess sess, *s;
	char buf[128], err[ERRMAX];

	memset(&sess, 0, sizeof(Sess));

	s = &sess;
	ARGBEGIN{
	case 'd':
		Debug = EARGF(usage());
		break;
	}ARGEND;

	quotefmtinstall();

	if(argc != 1)
		usage();

	s->dec = -1;
	s->webfs = "/mnt/web";

	for(i = 0; i < Nconn; i++){
		snprint(buf, sizeof(buf), "%s/clone", s->webfs);
		if((s->net[i] = open(buf, ORDWR)) < 0)
			sysfatal("%s cannot open - %r\n", buf);

		if((n = read(s->net[i], buf, sizeof buf-1)) < 0)
			sysfatal("%s cannot read - %r\n", buf);
		if(n == 0)
			sysfatal("%s is empty - %r\n", buf);
		buf[n] = '\0';
		s->conid[i] = atoi(buf);
	}

	setprio(getpid(), 16);

	/*
	 * if our connection drops, and we have found any valid
	 * url in the playlist we must retry. But we must first 
	 * fetch the playlist again, as it may have been updated.
	 * If nothing in the playlist was valid, give up.
	 */
	do{
		rc = play(s, 0, argv[0], 0, 0);
	}while(rc == 0);

	if(rc < 0){
		rerrstr(err, sizeof(err));
		print("error %q\n", err);
		exits(err);
	}

	exits(nil);
}
