#include <u.h>
#include <libc.h>
#include <bio.h>
#include "dat.h"
#include "fns.h"

int Debug;
int Nosound;

/*
 * If the user server name is a browser name (mozilla in it) then
 * you will get stats: e.g. http://stardust.wavestreamer.com:3353/
 *
 * add a semicolon to force the stream version of the data
 * or set the userserver name to "MPEG OVERRIDE"
 *
 * Strangely live365 seems to have problems if we use MPEG OVERRIDE
 * so I tried iTunes instead for a giggle.
 *
 * for V2 servers use:  http://stardust.wavestreamer.com:3353/;?sid=2
 * see
 * http://stackoverflow.com/questions/16871553/how-to-get-html5-to-play-a-stream-from-shoutcast
 *
 * We could start to experiment with shoutcast2 by setting Agent="Ultravox/2.1"
 */
char *Agent = "iTunes";

typedef struct Mime Mime;
static struct Mime {
	char *type;
	char *ext;
	int (*func)(Sess *s, Playlist *pl, int conn, char *url, int fd);
	char *decoder;
} Mimetypes[] = {
	{ "audio/x-scpls", 					"pls",		pls },
	{ "application/pls+xml", 			"pls",		pls },				/* vpr.net weirdness */
	{ "audio/x-mpegurl",				"m3u8",		m3u8 },				/* ascii */
	{ "application/x-mpegurl",			"m3u8",		m3u8 },				/* utf8 */
	{ "application/vnd.apple.mpegurl",	"m3u8",		m3u8 },				/* HLS stream */

	{ "audio/aacp", 			"aac",		nil,		"/bin/audio/adtsdec" },	/* AAC in MP2 TS */
	{ "audio/aac", 				"aac",		nil, 		"/bin/audio/adtsdec" },
	{ "audio/3gpp",				"3gpp",		nil, 		"/bin/audio/adtsdec" },
	{ "audio/3gpp2",			"3gpp2",	nil, 		"/bin/audio/adtsdec" },

	{ "audio/mpeg", 			"mpeg",		nil,		"/bin/audio/mad" },		/* MP1, MP2, or MP3 */
	{ "audio/x-mpeg", 			"mpeg",		nil,		"/bin/audio/mad" },
	{ "audio/mp2",				"mp2",		nil,		"/bin/audio/mad" },
	{ "audio/mp3",				"mp3",		nil,		"/bin/audio/mad" },

	{ "audio/x-aiff",			"aiff",		nil,		"/bin/audio/aiffdec" },

	{ "audio/mp4a-latm",		"m4a",		nil,		"/bin/audio/m4adec" },	/* AAC or ALAC in an MP4 wrapper */

	{ "audio/x-wav",			"wav",		nil,		"/bin/audio/wavdec" },

	{ "audio/x-flac",			"flac",		nil,		"/bin/audio/flacdec" },
	{ "application/ogg",		"ogg",		nil,		"/bin/audio/flacdec" },

	{ "video/MP2T", 			"ts",		nil,		"/bin/audio/mpgtsdec" },	/* HLS */

	{ nil, nil, nil }
};

static int
decompress(char *path)
{
	char *arg[10];
	int i, narg, fd[2];

	narg = 0;
	arg[narg++] = path;
	arg[narg++] = "-s";
	arg[narg++] = "-v";
	arg[narg] = nil;

	if(pipe(fd) < 0){
		fprint(2, "pipe failed %r");
		return -1;
	}

	switch(fork()){
	case -1:
		fprint(2, "fork failed %r");
		close(fd[0]);
		close(fd[1]);
		return -1;
	case 0:
		dup(fd[0], 0);
		close(fd[0]);
		close(fd[1]);
		close(1);
		for(i = 3; i < 63; i++)
			close(i);
		setprio(getpid(), 16);
		exec(path, arg);
		sysfatal("exec %s failed %r", path);
	default:
		close(fd[0]);
		break;
	}
	return fd[1];
}

static int
stream(Sess *s, int conn, int net, char *path)
{
	char buf[8192];
	int got, n, len, metaint;

	metaint = icyhdr(s, conn);
	if(Debug)
		fprint(2, "metaint: %d\n", metaint);

	if(strcmp(s->decpath, path) != 0){
		strcpy(s->decpath, path);
		if(s->dec != -1)
			close(s->dec);
		if((s->dec = decompress(path)) == -1)
			return -1;
	}

	if(metaint == -1){
		while(1){
			if((n = read(net, buf, sizeof(buf))) < 1)
				break;
			if(write(s->dec, buf, n) != n)
				break;
		}
	}
	else{
		while(1){
			for(got = 0; got < metaint; got += n){
	
				len = sizeof(buf);
				if(metaint - got < len)
					len = metaint - got;
	
				if((n = read(net, buf, len)) < 1)
					goto done;
				if(write(s->dec, buf, n) != n)
					goto done;
			}
			if(icyinband(net) < 0)
				break;
		}
	}

done:
	return 0;
}

static Mime *
mimetype(Sess *s, int conn, char *url)
{
	Mime *m;
	char *p, *ext, buf[1024];

	*buf = 0;
	if(httphdr(s, conn, "contenttype", buf, sizeof(buf)) == 0){
		if(Debug)
			fprint(2, "ContenType=%q\n", buf);

		/* some sites add other semicolon delimited fields */
		if((p = strchr(buf, ';')) != nil)
			*p = 0;

		for(m = Mimetypes; m->type; m++)
			if(cistrcmp(m->type, buf) == 0)
				return m;

		werrstr("%s unknown ContentType", buf);
	}
	else{	/* No contentype, try and match by extension (ugh!) */
		if((ext = strrchr(url, '.')) != nil){
			ext++;				/* step over '.' */
			if(Debug)
				fprint(2, "no ContenType, ext=%q\n", ext);

			for(m = Mimetypes; m->type; m++)
				if(cistrcmp(m->ext, ext) == 0)
					return m;
		}
		werrstr("%s no ContentType, no known header", buf);
	}

	return nil;
}

int
play(Sess *s, int conn, char *url)
{
	Mime *m;
	Chunk *ch;
	int fd, n, rc;
	Playlist plist;

	memset(&plist, 0, sizeof(Playlist));

	if((fd = httpdial(s, conn, Agent, url)) < 0){
		werrstr("cannot get playlist - %r");
		return -1;
	}

	if(Debug)
		fprint(2, "connected OK\n");

	if((m = mimetype(s, conn, url)) == nil){
		close(fd);
		return -1;
	}

	if(m->decoder){
		if(Debug)
			fprint(2, "%s audio stream\n", m->type);
		rc = stream(s, conn, fd, m->decoder);
		close(fd);
		return rc;
	}
	if(Debug)
		fprint(2, "%s playlist file\n", m->type);

	rc = (*m->func)(s, &plist, conn+1, url, fd);

	close(fd);

	if(s->seqno == 0)
		s->seqno = plist.seqno;

	for(ch = plist.chunks; ch; ch = ch->next){
		if(plist.seqno <= s->seqno){		/* only play new chunks */
			n = play(s, conn+1, ch->url);
			switch(n){
			case 0:
				break;
			case 1:
				if(rc == -1)
					rc = 0;
				rc++;
				break;
			case -1:
				rc = -1;
				break;
			}
		}
		s->seqno++;
		
	}

	freeplaylist(&plist);
	return rc;
}

static void
usage(void)
{
	fprint(2, "usage: %s [-n] audio-url [metadata-url]\n", argv0);
	fprint(2, "  -n   do not process audio, only display metadata\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int rc, n, i;
	Sess sess, *s;
	char buf[128], err[ERRMAX];

	memset(&sess, 0, sizeof(Sess));

	s = &sess;
	ARGBEGIN{
	case 'n':
		Nosound++;
		break;
	case 'd':
		Debug++;
		break;
	}ARGEND;

	quotefmtinstall();

//	if(argc != 1)
//		usage();

	s->dec = -1;
	s->webfs = "/mnt/web";

	for(i = 0; i < Nconn; i++){
		snprint(buf, sizeof(buf), "%s/clone", s->webfs);
		if((s->net[i] = open(buf, ORDWR)) < 0)
			sysfatal("%s cannot open - %r\n", buf);

		if((n = read(s->net[i], buf, sizeof buf-1)) < 0)
			sysfatal("%s cannot read - %r\n", buf);
		if(n == 0)
			sysfatal("%s is empty - %r\n", buf);
		buf[n] = '\0';
		s->conid[i] = atoi(buf);
	}

	setprio(getpid(), 16);

	/*
	 * if our connection drops, and we have found any valid
	 * url in the playlist we must retry. But we must first 
	 * fetch the playlist again, as it may have been updated.
	 * If nothing in the playlist was valid, give up.
	 */
	do{
		rc = play(s, 0, argv[0]);
	}while(rc >= 1);

	if(rc < 0){
		rerrstr(err, sizeof(err));
		fprint(2, "error %q\n", err);
		exits(err);
	}

	exits(nil);
}
