#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>
#include "dat.h"
#include "fns.h"

/*
 * Here we take care to read streams which have valid data on
 * their last line but are missing the last newline character.
 */
int
Bgetline(Biobuf *b, char *buf, int nbuf)
{
	int i, rv;
	Rune r;

	for(i = 0; i < nbuf - UTFmax - 1; ){
		rv = Bgetrune(b);
		if(rv == -1){
			if(i == 0)
				return -1;
			break;
		}
		if(rv == '\n')
			break;
		r = rv;
		i += runetochar(buf+i, &r);
	}
	if(i > 0 && buf[i-1] == '\r')
		buf[i-1] = 0;
	buf[i] = 0;
	return i;
}

int
setprio(int pid, int prio)
{
	int fd;
	char buf[128];

	snprint(buf, sizeof(buf), "/proc/%d/ctl", pid);
	if((fd = open(buf, OWRITE)) == -1)
		return -1;
	fprint(fd, "pri %d\n", prio);
	close(fd);
	return 0;
}

char *
skipws(char *p)
{
	while(isspace(*p))
		p++;
	return p;
}

int
httpredial(Sess *s, int conn, char *agent, char *url, char *etag)
{
	int e, n, fd, tries;
	char buf[128], *a[8], err[ERRMAX];

	tries = 0;
	if(Debug)
		fprint(2, "dialing conn=%d agent=%q url=%q\n", conn, agent, url);
Retry:
	snprint(buf, sizeof buf, "%s/%d/ctl", s->webfs, s->conid[conn]);
	if((fd = open(buf, OWRITE)) < 0){
		werrstr("cannot open another HTTP session - %r");
		return -1;
	}
	fprint(fd, "url %s", url);
	fprint(fd, "headers icy-metadata: 1");
	fprint(fd, "headers User-agent: %s", agent);
	if(etag && *etag)
		fprint(fd, "headers If-None-Match: %s", etag);
	close(fd);

	snprint(buf, sizeof buf, "%s/%d/body", s->webfs, s->conid[conn]);
	fd = open(buf, OREAD);

	if(fd < 0){
		rerrstr(err, sizeof(err));
		if(Debug)
			fprint(2, "connect failed err=%s\n", err);

		/*
		 * Wait whilst the socket closes, usually less than a second.
		 * We could use 2 sockets and ping pong between them but this
		 * is good enough for now.
		 */
		if(strstr(err, "client in use") != nil && tries++ < 10){
			sleep(300);
			goto Retry;
		}

		if((n = tokenize(err, a, nelem(a))) >= 3){
			e = atoi(a[1]);
			if(e == 0 && strcmp(a[n-1], "interrupted") == 0)
				werrstr("timeout on connect");
			else
				httperrstr(e);
		}
		return -1;
	}
	return fd;
}

int
httpdial(Sess *s, int conn, char *agent, char *url)
{
	return httpredial(s, conn, agent, url, nil);
}


int
httphdr(Sess *s, int conn, char *field, char *val, int len)
{
	int n, fd;
	char buf[128];

	if(val && len > 0)
		*val = 0;

	snprint(buf, sizeof(buf), "%s/%d/%s", s->webfs, s->conid[conn], field);
	if((fd = open(buf, OREAD)) < 0)
		return -1;
	n = read(fd, val, len-1);
	close(fd);

	if(n < 0)
		return -1;
	val[n] = 0;
	return 0;
}

void *
emalloc(long n)
{
	void *p;

	if((p = mallocz(n, 1)) == nil)
		sysfatal("No memory, %r");
	return p;
}

char *
estrdup(char *s)
{
	if((s = strdup(s)) == nil)
		sysfatal("No memory, %r");
	return s;
}

char *
esmprint(char *fmt, ...)
{
	char *s;
	va_list ap;

	va_start(ap, fmt);
	s = vsmprint(fmt, ap);
	va_end(ap);

	if(s == nil)
		sysfatal("No memory, %r");
	return s;
}

char *
absurl(char *base, char *url)
{
	char **m, *methods[] = { "http://", "ftp://", nil };

	for(m = methods; *m; m++)
		if(strncmp(*m, url, strlen(*m)) == 0)
			return estrdup(url);

	return esmprint("%s/%s", base, url);
}

double
iso2sec(char *s)
{
	Tm tm;
	double t, f;
	char *td[4], *ymd[5], *hms[5], *tz;

	/* e.g.
	 * 2010-04-06T11:39:46Z
	 * 2011-11-27t00:00:00.000+0000
	 */
	memset(&tm, 0, sizeof(tm));

	if(gettokens(s, td, nelem(td), "tT") != 2)
		return -1L;
	if(gettokens(td[0], ymd, nelem(ymd), "-") != 3)
		return -1L;
	if(gettokens(td[1], hms, nelem(hms), ":") != 2)

	tm.year = strtol(ymd[0], nil, 10) - 1900;
	tm.mon = strtol(ymd[1], nil, 10) -1;
	tm.mday = strtol(ymd[2], nil, 10);
	tm.hour = strtol(hms[0], nil, 10);
	tm.min = strtol(hms[1], nil, 10);
	tm.sec = strtol(hms[2], &tz, 10);
	
	f = 0;
	if(*tz == '.')
		f = strtod(tz, &tz);	/* fractional secconds */

	if(*tz == 'Z' || *tz == 'z')
		tm.tzoff = 0;
	else
	if(*tz == '+' || *tz == '-')
		tm.tzoff = strtol(tz, nil, 10);

	if((t = tm2sec(&tm)) < 0)
		return -1L;
	return t + f;
}

void
freeplaylist(Playlist *pl)
{
	Chunk *ch, *t;

	for(ch = pl->chunks; ch; ch = t){
		t = ch->next;
		free(ch->aux);
		free(ch->url);
		free(ch->title);
		free(ch);
	}
	free(pl->aux);
}
