#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>
#include "dat.h"
#include "fns.h"

int
dbg(char *s)
{
	if(Debug != nil && strstr(Debug, s) != nil)
		return 1;
	return 0;
}

/*
 * Here we take care to read streams which have valid data on
 * their last line but are missing the last newline character.
 */
int
Bgetline(Biobuf *b, char *buf, int nbuf)
{
	int i, rv;
	Rune r;

	for(i = 0; i < nbuf - UTFmax - 1; ){
		rv = Bgetrune(b);
		if(rv == -1){
			if(i == 0)
				return -1;
			break;
		}
		if(rv == '\n')
			break;
		r = rv;
		i += runetochar(buf+i, &r);
	}
	if(i > 0 && buf[i-1] == '\r')
		buf[i-1] = 0;
	buf[i] = 0;
	return i;
}

int
setprio(int pid, int prio)
{
	int fd;
	char buf[128];

	snprint(buf, sizeof(buf), "/proc/%d/ctl", pid);
	if((fd = open(buf, OWRITE)) == -1)
		return -1;
	fprint(fd, "pri %d\n", prio);
	close(fd);
	return 0;
}

char *
skipws(char *p)
{
	while(isspace(*p))
		p++;
	return p;
}

int
httpredial(Sess *s, int conn,  char *url, vlong start, vlong end, char *etag)
{
	int fd, tries;
	char buf[128], err[ERRMAX];

	tries = 0;
	if(dbg("http"))
		fprint(2, "dialing %lld->%lld %q\n", start, end, url);
Retry:
	snprint(buf, sizeof buf, "%s/%d/ctl", s->webfs, s->conid[conn]);
	if((fd = open(buf, OWRITE)) < 0){
		werrstr("cannot open another HTTP session - %r");
		return -1;
	}
	fprint(fd, "url %s", url);
	fprint(fd, "headers icy-metadata: 1");
	if(etag && *etag)
		fprint(fd, "headers If-None-Match: %s", etag);
	if(start < end)
		fprint(fd, "headers Range: bytes=%lld-%llud", start, end);
	close(fd);

	snprint(buf, sizeof buf, "%s/%d/body", s->webfs, s->conid[conn]);
	fd = open(buf, OREAD);

	if(fd < 0){
		rerrstr(err, sizeof(err));
		if(dbg("http"))
			fprint(2, "connect failed err=%s\n", err);

		/*
		 * Wait whilst the socket closes, usually less than a second.
		 * We could use 2 sockets and ping pong between them but this
		 * is good enough for now.
		 */
		if(strstr(err, "client in use") != nil && tries++ < 10){
fprint(2, "retry on conn\n");
			sleep(300);
			goto Retry;
		}

		httperrstr(err);
		return -1;
	}
	return fd;
}

int
httpdial(Sess *s, int conn, char *url, vlong start, vlong end)
{
	return httpredial(s, conn, url, start, end, nil);
}


int
httphdr(Sess *s, int conn, char *field, char *val, int len)
{
	int n, fd;
	char buf[128];

	if(val && len > 0)
		*val = 0;

	snprint(buf, sizeof(buf), "%s/%d/%s", s->webfs, s->conid[conn], field);
	if((fd = open(buf, OREAD)) < 0)
		return -1;
	n = read(fd, val, len-1);
	close(fd);

	if(n < 0)
		return -1;
	val[n] = 0;
	return 0;
}

void *
emalloc(long n)
{
	void *p;

	if((p = mallocz(n, 1)) == nil)
		sysfatal("No memory, %r");
	return p;
}

char *
estrdup(char *s)
{
	if((s = strdup(s)) == nil)
		sysfatal("No memory, %r");
	return s;
}

char *
esmprint(char *fmt, ...)
{
	char *s;
	va_list ap;

	va_start(ap, fmt);
	s = vsmprint(fmt, ap);
	va_end(ap);

	if(s == nil)
		sysfatal("No memory, %r");
	return s;
}

char *
absurl(char *base, char *url)
{
	char **m, *methods[] = { "http://", "ftp://", nil };

	for(m = methods; *m; m++)
		if(strncmp(*m, url, strlen(*m)) == 0)
			return estrdup(url);

	return esmprint("%s/%s", base, url);
}

char *
baseurl(char *url)
{
	char *p, *base;

	base = estrdup(url);
	if((p = strstr(base, "://")) == nil)
		sysfatal("%s bad url\n", url);
	p += 3;
	if((p = strrchr(p, '/')) == nil)
		sysfatal("%s bad url\n", url);
	*p = 0;
	return base;
}

void
freeplaylist(Playlist *pl)
{
	Chunk *ch, *t;

	for(ch = pl->chunks; ch; ch = t){
		t = ch->next;
		free(ch->aux);
		free(ch->url);
		free(ch->title);
		free(ch);
	}
	free(pl->aux);
	free(pl);
}
