#include <u.h>
#include <libc.h>
#include <thread.h>
#include <draw.h>
#include <keyboard.h>
#include <event.h>
#include <ctype.h>
#include <bio.h>

enum {			/* GUI layout constants */
	PAD = 5,
	MARGIN = 8,
	Chunk = 16,

	But1 = 1,
	But2 = 2,
	But3 = 4,
};

enum {			/* Fields, displayed in order so later items have priority */
	Mname,
	Mgenre,
	Mshow,
	Martist,
	Mtitle,
	Mrate,
	Mfreq,
	Mchans,
	Menc,
	Memph,
	Nmeta		/* Must be last */
};


struct {						/* map metadata keywords to GUI fields */
	char *s;
	int n;
} Map[] = {
	{ "name",	Mname },		/* from http headers */
	{ "genre",	Mgenre },

	{ "show",	Mshow },		/* from inband metadata or external helper program */
	{ "artist",	Martist },
	{ "title",	Mtitle },

	{ "enc",	Menc },			/* from audio decoder */
	{ "chans",	Mchans },
	{ "freq",	Mfreq },
	{ "rate",	Mrate },
	{ "emph",	Memph },

};


typedef struct Station Station;
struct Station {
	char *audiourl;		/* audio stream, .PLS, or .M4U URL */
	char *metaclass;	/* optional metadata class name */
	char *metaurl;		/* optional metadata source url */
	char *label;		/* what to display in the title field */
};

typedef struct Metadata Metadata;
struct Metadata {
	char *str;
	int dirty;
	Rectangle r;
};

typedef struct Volume Volume;
struct Volume {
	int l;
	int r;
};

enum {						/* GUI layout, widths in chars */
	Wcol1 = 20, 			/* width of collum 1 */
	Wstation = 32,			/* width of station name */
	Wgenre = 32,			/* width of genre */
	Wtitle = 80,			/* width of title, artist, source */

	Horz = 10, 				/* gap between vertical elements */
	Vert = 8 				/* gap between horizontal elements */
};

struct {					/* cached data, for compression ratio calculation */
	int depth;
	int nchans;
	double rate;
	double freq;
} Cache;

typedef struct Cmd Cmd;
struct Cmd {
	int		slot;			/* slot in radio station table */
	int		fds[2];			/* file descriptors in pipe */
	Channel	*cpid;			/* channel to receive new pid */
	int		pid;			/* pid of command */
};

Volume Vol;					/* current volume level */
Metadata Meta[Nmeta];		/* metadata read from radio player */
Cmd Current;				/* currently selected station */
Station *Stations;			/* known station list */
int Nstations;				/* size of above array */
Font *Face;					/* current font */
Image *Background;			/* background colour */
char *Stationlist;			/* filename of station list */
int Debug;					/* chatter */
int Fixname;				/* don't update station name from metadata */

char *Radio = "/bin/audio/radio";

void*
erealloc(void *v, ulong n)
{
	v = realloc(v, n);
	if(v == nil)
		sysfatal("out of memory reallocating %lud", n);
	return v;
}

void*
emalloc(ulong n)
{
	void *v;

	v = malloc(n);
	if(v == nil)
		sysfatal("out of memory allocating %lud", n);
	memset(v, 0, n);
	return v;
}

char*
estrdup(char *s)
{
	int l;
	char *t;

	if (s == nil)
		return nil;
	l = strlen(s)+1;
	t = emalloc(l);
	memcpy(t, s, l);

	return t;
}

char *
trim(char *s)
{
	char *p;

	for(p = strrchr(s, 0) -1; p >= s && isspace(*p); p--)
		*p = 0;
	for(p = s; isspace(*p); p++)
		continue;
	return p;
}

void
readlist(void)
{
	int n, m;
	Biobuf *bp;
	char *line, *a[32];

	if((bp = Bopen(Stationlist, OREAD)) == nil)
		return;

	m = 0;
	Nstations = 0;
	while((line = Brdline(bp, '\n')) != nil){
		line[Blinelen(bp)-1] = 0;
		n = tokenize(line, a, nelem(a));
		if(n < 2)				/* blank/short lines */
			continue;
		if(a[0][0] == '#')		/* comment */
			continue;
		if(Nstations >= m){
			m += Chunk;
			Stations = erealloc(Stations, (m+Chunk) * sizeof(Station));
		}

		Stations[Nstations].label = estrdup(a[0]);
		Stations[Nstations].audiourl = estrdup(a[1]);

		Stations[Nstations].metaclass = nil;
		if(n > 2)
			Stations[Nstations].metaclass = estrdup(a[2]);

		Stations[Nstations].metaurl = nil;
		if(n > 3)
			Stations[Nstations].metaurl = estrdup(a[3]);
		Nstations++;
	}
	Bterm(bp);
}

void
readvolume(Volume *v)
{
	Biobuf *bp;
	char *line, *a[4];

	if((bp = Bopen("/dev/volume", OREAD)) == nil)
		return;
	while((line = Brdline(bp, '\n')) != nil){
		line[Blinelen(bp)-1] = 0;
		if(tokenize(line, a, nelem(a)) < 3)
			continue;
		if(strcmp(a[0], "audio") == 0){
			v->l = atoi(a[1]);
			v->r = atoi(a[2]);
		}
	}
	Bterm(bp);
}

void
writevolume(Volume *v)
{
	int fd;

	if(v->l < 0 || v->r < 0)
		v->l = v->r = 0;
	if(v->l > 100 || v->r > 100)
		v->l = v->r = 100;

	if((fd = open("/dev/volume", OWRITE)) == -1){
		fprint(2, "%s: /dev/volume cannot open\n", argv0);
		return;
	}
	fprint(fd, "audio %d %d\n", v->l, v->r);
	close(fd);
}

void
set(int fld, char *val)
{
	free(Meta[fld].str);
	Meta[fld].str = estrdup(trim(val));
	Meta[fld].dirty = 1;
}


void
updfreq(void)
{
	if(Cache.freq > 80 && Cache.freq < 100)
		set(Mfreq, "double");
	if(Cache.freq > 40 && Cache.freq < 50)
		set(Mfreq, "full");
	if(Cache.freq > 20 && Cache.freq < 25)
		set(Mfreq, "half");
	if(Cache.freq >  10 && Cache.freq < 13)
		set(Mfreq, "quarter");
}

void
updrate(void)
{
	char *buf;
	double ratio;

	if(Cache.rate == 0)
		return;
	ratio = (Cache.freq*1000*Cache.nchans*Cache.depth) / (Cache.rate*1024) + 0.5;
	buf = smprint("%.0f:1", ratio);
	set(Mrate, buf);
	free(buf);
}

void
readmeta(Channel *ch)
{
	int i, m;
	char *line, *a[6];

	while((line = recvp(ch)) != nil){
		if(tokenize(line, a, nelem(a)) < 2)
			continue;

		if(strcmp(a[0], "error") == 0)
			set(Mtitle, a[1]);

		for(i = 0; i < nelem(Map); i++){
			if(strcmp(a[0], Map[i].s) == 0){
				m = Map[i].n;

				if(Fixname && m == Mname)
					continue;

				set(m, a[1]);
				switch(m){
				case Mfreq:
					Cache.freq = atof(a[1]);
					updfreq();
					updrate();
					break;
				case Mrate:
					Cache.rate = atof(a[1]);
					updrate();
					break;
				case Mchans:
					if(strstr(a[1], "Mono") != nil)
						Cache.nchans = 1;
					else
					if(strstr(a[1], "Stereo") != nil || strstr(a[1], "Joint") != nil)
						Cache.nchans = 2;
					else
						Cache.nchans = atoi(a[1]);
					updrate();
					break;
				default:
					break;
				}
			}
		}
		free(line);
	}
}

void
drawmeta(int i)
{
	draw(screen, Meta[i].r, Background, nil, ZP);
	if(Meta[i].str)
		_string(screen, addpt(Meta[i].r.min, Pt(2,0)), display->black, ZP,
			Face, Meta[i].str, nil, strlen(Meta[i].str), 
			Meta[i].r, nil, ZP, SoverD);
	if(Debug)
		border(screen, Meta[i].r, 1, display->black, ZP);	
	Meta[i].dirty = 0;
}

Rectangle 
field(Point p, Point fs, int n)
{
	if(n < 0)
		return Rpt(addpt(screen->r.min, p), addpt(screen->r.min, Pt(screen->r.max.x-Horz, p.y+fs.y)));
	else
		return Rpt(addpt(screen->r.min, p), addpt(screen->r.min, Pt(p.x+fs.x*n, p.y+fs.y)));
}

int
geometry(void)
{
	Point p1, p2, p3;
	static Rectangle scr;
	Point fs = stringsize(Face, "j");

	if(eqrect(scr, screen->r))
		return 0;
	scr = screen->r;

	p1 = Pt(Horz, Vert);
	Meta[Mchans].r = field(p1, fs, Wcol1);
	p2 = Pt(p1.x+Dx(Meta[Mchans].r), p1.y);
	Meta[Mname].r = field(p2, fs, Wstation);
	p3 = Pt(p2.x+Dx(Meta[Mname].r), p1.y);
	Meta[Mgenre].r = field(p3, fs, -1);

	p1 = Pt(p1.x, p1.y+Dy(Meta[Mchans].r)+Vert/2);
	Meta[Mfreq].r = field(p1, fs, Wcol1);

	p1 = Pt(p1.x, p1.y+Dy(Meta[Mfreq].r)+Vert/2);
	Meta[Menc].r = field(p1, fs, Wcol1);
	p2 = Pt(p1.x+Dx(Meta[Menc].r), p1.y-Vert);
	Meta[Mshow].r = field(p2, fs, -Wtitle);

	p1 = Pt(p1.x, p1.y+Dy(Meta[Menc].r)+Vert/2);
	Meta[Mrate].r = field(p1, fs, Wcol1);
	p2 = Pt(p1.x+Dx(Meta[Mrate].r), p1.y);
	Meta[Martist].r = field(p2, fs, -Wtitle);

	p1 = Pt(p1.x, p1.y+Dy(Meta[Mrate].r)+Vert/2);
	Meta[Memph].r = field(p1, fs, Wcol1);
	p2 = Pt(p1.x+Dx(Meta[Memph].r), p1.y);
	Meta[Mtitle].r = field(p2, fs, -Wtitle);

	return 1;
}

void
clr(char *s)
{
	Cache.depth = 16;
	Cache.nchans = 2;
	Cache.rate = 2.;
	Cache.freq = 2;

	set(Martist, "");
	set(Mchans, "");
	set(Menc, "");
	set(Mfreq, "");
	set(Mgenre, "");
	set(Mshow, "");
	set(Mrate, "");
	set(Mtitle, "");
	set(Memph, "");
	set(Mname, s);
}

void
redraw(Image *screen, int all)
{
	int i;

	all |= geometry();
	if(all)
		draw(screen, screen->r, Background, nil, ZP);

	for(i = 0; i < Nmeta; i++)
		if(all || Meta[i].dirty)
			drawmeta(i);
}

void
eresized(int new)
{
	if(new && getwindow(display, Refmesg) < 0)
		fprint(2,"can't reattach to Stationsdow");
	geometry();
	redraw(screen, 1);
}

static void
radioproc(void *v)
{
	int n;
	Cmd *c;
	char *args[8];
	
	threadsetname("radioproc");
	
	r = v;
	dup(c->fds[1], 1);
	close(c->fds[0]);
	close(c->fds[1]);
	
	n = 0;
	args[n++] = Stations[c->slot].audiocmd;
	args[n++] = Stations[c->slot].audiourl;
	args[n] = nil;

	procexec(r->cpid, c, args);
	sysfatal("%s exec failed %r", c->cmd);
}

static void
metaproc(void *v)
{
	int n;
	Cmd *c;
	char *args[8];
	
	threadsetname("metaproc");
	
	r = v;
	dup(c->fds[1], 1);
	close(c->fds[0]);
	close(c->fds[1]);
	
	n = 0;
	args[n++] = Stations[c->slot].metacmd;
	args[n++] = Stations[c->slot].metaurl;
	args[n] = nil;

	procexec(r->cpid, c, args);
	sysfatal("%s exec failed %r", c->cmd);
}

void
newstation(Cmd *r)
{
	threadsetname("execproc");

	clr(Stations[r->slot].label);

	if(r->audiopid > 0)
		postnote(PNGROUP, r->audiopid, "kill");

	if(pipe(r->fds) < 0)
		sysfatal("pipe: %r");
	procrfork(radioproc, r, 8192, RFFDG);
	close(r->fds[1]);
	return r->fd[0];
}

char *
labgen(int i)
{
	if(i < 0 || i >= Nstations)
		return nil;
	return Stations[i].label;
}

void
stationsel(Mouse *m)
{
	static Menu men = { .gen = labgen };
	int hit;

	men.lasthit = Current.slot;
	hit = emenuhit(3, m, &men);
	if(hit == -1 || hit == Current.slot)
		return;
	Current.slot = hit;
	newstation(&Current);
}

void
bumpstation(int incr)
{
	if(Current.slot + incr < 0 || Current.slot >= Nstations)
		return;
	Current.slot += incr;
	newstation(&Current);
}

void
ctrl(Mouse *m)
{
	int hit;
	static char *item[] = { "Vol+", "Vol-", nil };
	static Menu men = { .item = item };

	hit = emenuhit(2, m, &men);
	if(hit == -1)
		return;

	switch(hit){
	case 0:
		Vol.l += 5;
		Vol.r += 5;
		break;
	case 1:
		Vol.l -= 5;
		Vol.r -= 5;
		break;
	default:
		return;
	}

	writevolume(&Vol);
}
	

void
usage(void)
{
	fprint(2, "usage: tune [-f font] [-s] [-n]\n");
	threadexits("usage");
}


void
tuner(void)
{
	Mouse mm;
	Mousectl *mc;
	Keyboardctl *kc;
	Rune km;

	Alt a[] = {
		/* c	v	op */
		{nil,	&dm,	CHANRCV},	/* data from stdin */
		{nil,	&mm,	CHANRCV},	/* mouse message */
		{nil,	&km,	CHANRCV},	/* keyboard runes */
		{nil,	&rm,	CHANRCV},	/* resize event */
		{nil,	nil,	CHANEND},
	};

	memset(&mm, 0, sizeof mm);
	memset(&km, 0, sizeof km);
	dm = rm = 0;

	Current.pid = -1;
	Current.slot = -1;

	einit(Emouse|Ekeyboard);
	Etimer = etimer(0, 2500);

	if((dc = chancreate(sizeof dm, 10)) == nil)
		sysfatal("chancreate: %r");
	if((mc = initmouse(nil, screen)) == nil)
		sysfatal("initmouse: %r");
	if((kc = initkeyboard(nil)) == nil)
		sysfatal("initkeyboard: %r");

	a[0].c = dc;
	a[1].c = mc->c;
	a[2].c = kc->c;
	a[3].c = mc->resizec;

	proccreate(reader, a[0].c, STACK + sizeof(Biobuf));

	for(;;){
		switch(alt(a)){
		case 0:
			break;
		case 1:
			if(e.mouse.buttons == But2)
				ctrl(&e.mouse);
			if(e.mouse.buttons == But3)
				stationsel(&e.mouse);
			break;
		case 2:
			switch(km){
			case 0x7F:
				if(Pid > 0)
					postnote(PNGROUP, Pid, "kill");
				threadexits(0);
				break;
			case Kup:
				bumpstation(-1);
				break;
			case Kdown:
				bumpstation(1);
				break;
			default:
				break;
			}
			break;
		case 3:
			redraw(screen, 0);
			break;
		default:
			sysfatal("shouldn't happen");
		}
		redraw(screen, 0);
	}
}

int threadmainstacksize = 1024 * 64;

void
threadmain(int argc, char **argv)
{
	char *home, *fontname;

	fontname = "/lib/font/bit/lucidasans/unicode.8.font";
	ARGBEGIN{
	case 'd':
		Debug++;
		break;
	case 'f':
		fontname = EARGF(usage());
		break;
	case 'n':
		Fixname = 1;
		break;
	default:
		usage();
	}ARGEND

	if(argc)
		usage();

	if((home = getenv("home")) == nil)
		sysfatal("$home not set\n");
	if((Stationlist = snprint("/%s/lib/radio", home)) == nil)
		sysfatal("No memory\n");
	free(home);

	initdraw(0, 0, "Stationswatch");
	Background = allocimagemix(display, DPalebluegreen, DWhite);
	if(Background == nil)
		sysfatal("allocimagemix: %r");
	if((Face = openfont(display, fontname)) == nil)
		sysfatal("%s not found", fontname);

	clr("Nothing slected");

	Vol.l = Vol.r = 10;
	writevolume(&Vol);
	readvolume(&Vol);

	readlist();
	redraw(screen, 1);
	tuner();

}
