#include <u.h>
#include <libc.h>
#include <bio.h>

enum {
	FmtPCM = 1,
	FmtAdpcm = 2,
	FmtFloat = 3,
	FmtAlaw = 6,
	FmtMulaw = 7,
	FmtIma = 17,
	FmtG723 = 20,
	FmtGSM = 49,
	FmtG721 = 64,
	FmtMPEG = 80,
	FmtExt = 0xfffe,

	MaxDataSz = 4294966990,
};

void
wr(Biobuf *bp, int width, uint val)
{
	int i;

	for(i = 0; i < width; i++){
		if(Bputc(bp, val & 0xff) < 0)
			sysfatal("write failed %r");
		val >>= 8;
	}
}

void
pr(Biobuf *bp, char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	if(Bvprint(bp, fmt, ap) <= 0)
		sysfatal("write failed %r");
	va_end(ap);
}

void
wavhdr(Biobuf *bp, int rate, int chans, int depth, vlong sz)
{
    uint bytes;

	bytes = (depth + 7) / 8;
	if(sz < 0 || sz > MaxDataSz)
		sz = MaxDataSz;
	sz += 44;

	pr(bp, "RIFF");					/* chunk ID */
	wr(bp, 4, sz -8);				/* size of following  */
	pr(bp, "WAVE");					/* format ID */

	pr(bp, "fmt ");					/* subchunk ID */
	wr(bp, 4, 16);					/* subchunk size */
	wr(bp, 2, FmtPCM);				/* audio format */ 
	wr(bp, 2, chans);				/* number of channels */
	wr(bp, 4, rate);				/* sample rate */
	wr(bp, 4, rate*chans*bytes);	/* byte rate */
	wr(bp, 2, chans*bytes);			/* block align */
	wr(bp, 2, depth);				/* bits per sample */

	pr(bp, "data");					/* chunk ID */
	wr(bp, 4, sz);
}

static void
usage(void)
{
	fprint(2, "%s [file]\n", argv0);
	exits("usage");
}

void
main(int argc, char *argv[])
{
	vlong sz;
	char buf[8192];
	Biobuf out, inp, *bp;
	int n, rate, chans, depth;

	chans = 2;
	depth = 16;
	rate = 44100;
	ARGBEGIN{
	case 'r':
		rate = atoi(EARGF(usage()));
		break;
	case 'c':
		chans = atoi(EARGF(usage()));
		break;
	case 'd':
		depth = atoi(EARGF(usage()));
		break;
	default:
		usage();
		break;
	}ARGEND;

	if(argc > 1)
		usage();
	if(rate < 4000 || rate > 96000)
		sysfatal("%d illegal sample rate, must be between 4000 and 96000\n", rate);
	if(chans < 1 || chans > 8)
		sysfatal("%d illegal channel count, must be between 1 and 8\n", chans);
	if(depth < 8 || depth > 32)
		sysfatal("%d illegal depth, must be between 8 and 32\n", depth);

	Binit(&out, 1, OWRITE);

	if(argc == 0){
		Binit(&inp, 0, OREAD);
		bp = &inp;
	}
	else{
		if((bp = Bopen(argv[0], OREAD)) == nil)
			sysfatal("%s cannot open %r\n", argv[0]);
	}

	/*
	 * dirfstat() is not apropriate as a pipe
	 * will tell us how much is buffered not the
	 * data size. if seek fails then we just set
	 * the size to -1 (as big as possible).
	 */
	sz = Bseek(bp, 0, 2);
	Bseek(bp, 0, 0);

	wavhdr(&out, rate, chans, depth, sz);

	while((n = Bread(bp, buf, sizeof(buf))) > 0)
		if(Bwrite(&out, buf, n) < 0)				/* FIXME: force byte order, the world is not intel */
			sysfatal("write error %r\n");
	exits(nil);
}
